//  BRAT is a whole genome bisulfite sequence mapping program
//  Copyright (C) 2009  Elena Yavorska Harris
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include<iomanip>
#include<map>
#include<bitset>
#include<ctime>
#include<list>
#include<cmath>
#include<vector>
#include<string>
#include<iostream>
#include<fstream>
#include<sstream>
#include<algorithm>
using namespace std;

class Point
{

public:
	//constructors of this class
	Point(long int xx, long int yy) : x(xx), y(yy) {};
	Point() : x(0), y(0) {};
	
	
	void operator =(Point a) { x = a.x; y = a.y;};
	//x is index, y is tails
	bool operator <(const Point& a) const
	{ 
		if(y < a.y) return true; 
		else return false;
	}


	long int x;
	long int y;

};
class Cov{
public:
	Cov(unsigned short int x, unsigned short int y, unsigned short int z, unsigned short int w) : a(x), c(y), g(z), t(w) {};
	Cov() : a(0), c(0), g(0), t(0) {};

	void operator =(Cov s){ a = s.a; c = s.c; g = s.g; t = s.t; };
	unsigned short int a;
	unsigned short int c;
	unsigned short int g;
	unsigned short int t;
};
long int str_to_int(string s)
{
	istringstream is(s);
	long int result = -1;
	is >> result;
	return result;
}
double str_to_double(string s)
{
	istringstream is(s);
	double result = -1;
	is >> result;
	return result;
}

string int_to_str(long int x)
{
	stringstream outstr;
	outstr << x;
	return outstr.str();

}
string double_to_str(double x)
{
	stringstream outstr;
	outstr << x;
	return outstr.str();

}

long int max(long int x, long int y)
{	
	if(x > y)
		return x;
	else
		return y;

}
long int min(long int x, long int y){ 
	if(x < y)
		return x;
	else
		return y;
}
long int rand_res(long int total_len, double i)
{
	srand( time(0)*i );
	long int res = 0 + rand() % (total_len - 1);
	return res;
}

char rev(char x)
{
	if(x == 'a')
		return 'c';
	else if(x == 'c')
		return 'g';
	else
		return 'a';

}//rev

double choose(double top, double bottom){
	//given top and bottom it returns Log_10(choose(top, bottom)), where choose is binomial coef
	if(bottom == 0 || bottom == top)
		return 0; //choose = 1, log(choose)=0
	else if(bottom == 1)
		return log10(top);
	double sum = 0.0, current;
	long int i, j;

	long int stop = top - bottom;
	for(i = top; i > stop; i--){
		current = log10(top) - log10(bottom);
		sum += current;
		top -= 1.0;
		bottom -= 1.0;
		
	}//for
	
	 return sum; 
	
}//choose
double p_value(double total_genes, double cluster, double colors, double total_all){
	/*
	Given a set of total_genes' sequences, 
	given a set of chosen genes (cluster) sequences,
	given # occurrences of motif in cluster,
	given # occurrences of motif in all_genes
	
	Find hypergeometric log_10(p_value)
	
	*/
	double stop = total_all;
	if(cluster < total_all)
		stop = cluster;
	double N = total_genes;
	double N_cluster = N - cluster;
	
	long int i;
	double j;//total_all - i
	double sum = 0.0, cluster_choose_i, N_cluster_choose_j , N_choose_total_all, current;
	for(i = colors; i <= stop; i++){
		j = total_all - i;
		cluster_choose_i = choose(cluster, i);
		N_cluster_choose_j = choose(N_cluster, j);
		N_choose_total_all = choose(N, total_all);
		current = (cluster_choose_i + N_cluster_choose_j - N_choose_total_all);// 
		sum += pow(10.0, current);
	}
	
	
	return (-log10(sum));
//return sum;	
}
void some(){



}
string rev_complement(string s)
{
	string res("");
	int alen = s.length() - 1; 
	for(int j = alen; j >= 0; j--)
{
		if(s[j] == 'A')
			res = res + 'T';
		else if(s[j] == 'T')
			res = res + 'A';
		else if(s[j] == 't')
			res = res + 'a';
		else if(s[j] == 'a')
			res = res + 't';
		else if(s[j] == 'C')
			res = res + 'G';
		else if(s[j] == 'G')
			res = res + 'C';
		else if(s[j] == 'c')
			res = res + 'g';
		else if(s[j] == 'g')
			res = res + 'c';
		else
			res = res + 'N';
		
	}
	return res;

}//rev_comple

void parse_options(int argc, char* argv[], string &pairs, 
				   string &singles, string &pref, string &genome)
{
	long int i, j;
	for(i = 1; i < argc; i++){
		if(argv[i][0] != '-'){
			cerr << "\nERROR: Invalid char=" << argv[i][0] << "=this" << endl ;		
			return;
		}
		switch(argv[i][1]){
		case 'p': pairs = argv[++i]; break;
		case 's': singles = argv[++i]; break;
		case 'P': pref = argv[++i]; break;
		case 'r': genome = argv[++i]; break;
		default: cerr << "ERROR: Invalid char=" << argv[i][1] << "=this" << endl ;  exit(0); break;
		}
	}//for i

}//parse_opt()


int main(int argc, char* argv[]){

	srand( time(0) );

	string read1, read2, strand;
	long int id, i, j, asize, u, y, chrom_id, pos1, pos2, mism, mism1, mism2;
	long int count = 0;

	bool is_pe = false;
	string pairs(""), singles(""), pref, references;
	parse_options(argc, argv, pairs, singles, pref, references);

	if(pairs.length() > 0)
		is_pe = true;

	ifstream in;
	in.open(references.c_str(), ios::in);
	if(!in){
		cerr << "\nERROR: cannot open " << references << endl;
		exit(0);
	}
	vector<string> gen_names;
	in >> read1;
	while(!in.eof()){
		gen_names.push_back(read1);
		in >> read1;
	}//while
	in.close(); in.clear();

	vector<long int> size_chrom;
	
	asize = gen_names.size();
	string line;
	map<string, long int> chrom_names;
	vector<string> chroms_names;
		string chrom_name ;

	for(i = 0; i < asize; i++){
	
		in.open(gen_names[i].c_str(), ios::in);

		if(!in){
			cerr << "\nERROR: cannot open " << gen_names[i] << endl;
			exit(0);
		}//if
cout << "opening " << gen_names[i] << endl;
		getline(in, line);//first line of fasta file
		istringstream istr(line);
		char ch;
		istr >> ch;
		istr >> chrom_name;
		chrom_names.insert(make_pair(chrom_name, i));
		chroms_names.push_back(chrom_name);

		getline(in, line);
		long int cur_size = 0;
		while(!in.eof()){
			cur_size += line.length();
			getline(in, line);
		}//while
		in.close(); in.clear();

		size_chrom.push_back(cur_size);
cout << "size of " << chrom_name << " is " << cur_size << endl;
	}//for

	long int num_chroms = size_chrom.size();

	//to save space, we process references sequentially
	long int thresh_size = 15000000;
	vector< map<string, Point> > ref_seq;
	map<string, Point> first_map;
	ref_seq.push_back(first_map);
	long int last_map_ind = 0;
	long int cur_index = 0;
	long int cur_size = 0;

cout << "current map: " ;
	for(i = 0; i < num_chroms; i++){
		cur_size += size_chrom[i];
		ref_seq[last_map_ind].insert(make_pair(chroms_names[i], Point(cur_index, size_chrom[i])));
		cur_index++;
		cout << chroms_names[i] << ", " ;
		if(cur_size >= thresh_size && i < num_chroms-1){
			map<string, Point> amap;
			ref_seq.push_back(amap);
			last_map_ind++;
			cur_size = 0;
			cur_index = 0;
cout << endl;
cout << "current map: " ;
		}//if
	
	}//for i
cout << endl;


	vector<string> names;
	vector<string> names_singles2;


	if(is_pe == true){
		string aname;
		in.open(pairs.c_str(), ios::in);
		if(!in){
			cerr << "ERROR: can't open " << pairs << endl;
			exit(0);
		}
		in >> aname;
		while(!in.eof()){
			names.push_back(aname);
			in >> aname;
		}//while
		in.close(); in.clear();
cout << "the number of files with pair-end results is " << names.size() << endl;
	}//if

	long int chrom_num = size_chrom.size();
	string aname;
	bool is_singles = false;
	if(singles.length() > 0){
		is_singles = true;
	}
	if(is_singles == true){
		in.open(singles.c_str(), ios::in);
		if(!in){
			cerr << "ERROR: can't open " << singles << endl;
			exit(0);
		}
		in >> aname;
		while(!in.eof()){
			names_singles2.push_back(aname);
			in >> aname;
		}//while
		in.close(); in.clear();

cout << "the number of files with single-end results is " << names_singles2.size() << endl;

	}//if singles

	long int t;
	cur_size = ref_seq.size();
	long int amap_size ;
	for(t = 0; t < cur_size; t++){

		amap_size = ref_seq[t].size();

		vector<long int> sizes_chroms_cur_map(amap_size);
		map<string, Point>::iterator start = ref_seq[t].begin();
		map<string, Point>::iterator end = ref_seq[t].end();
		for(; start != end; start++){
			long int ind = (start->second).x;
			long int asize_cur = (start->second).y;
			sizes_chroms_cur_map[ind] = asize_cur;
		}//for

		vector< vector<Cov> > genome;
		vector< vector<Cov> > rev_genome;
		for(i = 0; i < amap_size; i++){
			vector<Cov> dum(sizes_chroms_cur_map[i]);
			genome.push_back(dum);
			rev_genome.push_back(dum);
		}//for i

		if(is_pe == true){

cout << "Processing Pair-End results" << endl;
			for( y = 0; y < names.size(); y++){


				in.open(names[y].c_str(), ios::in);
				if(!in){
					cerr << "ERROR: can't open " << names[y] << endl;
					exit(0);
				}
				else
					cout << "opening " << names[y] << endl;

				long int count_lines = 0;
				in >> id;


				while(!in.eof()){
					count_lines++;
					in >> read1 ;
					in >> read2;
					in >> chrom_name; //starts with 1

					in >> strand;
					in >> pos1;
					in >> pos2;
					in >> mism1;
					in >> mism2;

					map<string, Point>::iterator afind = ref_seq[t].find(chrom_name);

					if(afind != end){

						chrom_id = (afind->second).x ;
						asize = (afind->second).y;

						if(chrom_id < 0 || chrom_id >= chrom_num)
						{
						cout << "chrom_id is out of boundary at line " << count_lines << endl;
						exit(0);

						}			
		

						long int len_mer1 = read1.length();

						long int len_mer2 = read2.length();
						j = 0;

						long int forw_len = len_mer1;
						long int rev_len = len_mer2;

						string forw_read = read1;
						string rev_read = read2;
						long int forw_pos = pos1;
						long int rev_pos = pos2;
						if(strand == "-"){
							forw_read = read2;
							rev_read = read1;
							forw_pos = pos2;
							rev_pos = pos1;
							forw_len = len_mer2;
							rev_len = len_mer1;
						}

						if(strand == "+"){

						for(i = forw_pos; i < forw_pos + forw_len; i++){

							if(i >= 0 && i < asize){
								if(forw_read[j] == 'A' || forw_read[j] == 'a')
									genome[chrom_id][i].a++;
								else if(forw_read[j] == 'C' || forw_read[j] == 'c')
									genome[chrom_id][i].c++;
								else if(forw_read[j] == 'G' || forw_read[j] == 'g')
									genome[chrom_id][i].g++;
								else if(forw_read[j] == 'T' || forw_read[j] == 't')
									genome[chrom_id][i].t++;
							}//if valid position
							j++;
						}//for i

						}//if first mate is mapped to positive strand 
						else{
							for(i = forw_pos; i < forw_pos + forw_len; i++){

								if(i >= 0 && i < asize){
									if(forw_read[j] == 'A' || forw_read[j] == 'a')
										rev_genome[chrom_id][i].a++;
									else if(forw_read[j] == 'C' || forw_read[j] == 'c')
										rev_genome[chrom_id][i].c++;
									else if(forw_read[j] == 'G' || forw_read[j] == 'g')
										rev_genome[chrom_id][i].g++;
									else if(forw_read[j] == 'T' || forw_read[j] == 't')
										rev_genome[chrom_id][i].t++;
								}//if valid position
								j++;
							}//for i						
						}//else rev-complement of first map is mapped to positive strand
						//process reverse read

						//NOTE: we count acgt as they map forward strand,
						//so if we need to use this count for reverse strand, 
						//we have to switch counts of g and c AND counts of t and a
						//check if there is an overlap between two mates
						bool overlap = false;
						if(rev_pos < forw_pos + forw_len)
							overlap = true;
						if(overlap == false){//no overlap

						j=rev_len -1;
						if(strand == "-"){
							for(i = rev_pos; i < rev_pos + rev_len; i++){

								if(i >= 0 && i < asize){


									if(rev_read[j] == 'A' || rev_read[j] == 'a')
										rev_genome[chrom_id][i].t++;
									else if(rev_read[j] == 'C' || rev_read[j] == 'c')
										rev_genome[chrom_id][i].g++;
									else if(rev_read[j] == 'G' || rev_read[j] == 'g')
										rev_genome[chrom_id][i].c++;
									else if(rev_read[j] == 'T' || rev_read[j] == 't')
										rev_genome[chrom_id][i].a++;
								}
								j--;
							}//for i
						}//if strand -
						else{
							for(i = rev_pos; i < rev_pos + rev_len; i++){

								if(i >= 0 && i < asize){


									if(rev_read[j] == 'A' || rev_read[j] == 'a')
										genome[chrom_id][i].t++;
									else if(rev_read[j] == 'C' || rev_read[j] == 'c')
										genome[chrom_id][i].g++;
									else if(rev_read[j] == 'G' || rev_read[j] == 'g')
										genome[chrom_id][i].c++;
									else if(rev_read[j] == 'T' || rev_read[j] == 't')
										genome[chrom_id][i].a++;
								}
								j--;
							}//for i
						}//else strand +
						}//no overlap
						else{//overlap
							long int pos_overlapped = forw_pos + forw_len - rev_pos; 
							long int rev_stop = rev_pos + rev_len;
                                                        rev_pos = forw_pos + forw_len;
							
                                                j=rev_len -1 - pos_overlapped;

												if(strand == "-"){

                                                for(i = rev_pos; i < rev_stop; i++){

                                                        if(i >= 0 && i < asize){


                                                                if(rev_read[j] == 'A' || rev_read[j] == 'a')
                                                                        rev_genome[chrom_id][i].t++;
                                                                else if(rev_read[j] == 'C' || rev_read[j] == 'c')
                                                                        rev_genome[chrom_id][i].g++;
                                                                else if(rev_read[j] == 'G' || rev_read[j] == 'g')
                                                                        rev_genome[chrom_id][i].c++;
                                                                else if(rev_read[j] == 'T' || rev_read[j] == 't')
                                                                        rev_genome[chrom_id][i].a++;
                                                        }
                                                        j--;
                                                }//for i
												}//if strand -
												else{
													for(i = rev_pos; i < rev_stop; i++){

															if(i >= 0 && i < asize){


																	if(rev_read[j] == 'A' || rev_read[j] == 'a')
																			genome[chrom_id][i].t++;
																	else if(rev_read[j] == 'C' || rev_read[j] == 'c')
																			genome[chrom_id][i].g++;
																	else if(rev_read[j] == 'G' || rev_read[j] == 'g')
																			genome[chrom_id][i].c++;
																	else if(rev_read[j] == 'T' || rev_read[j] == 't')
																			genome[chrom_id][i].a++;
															}
															j--;
													}//for i												
												}//if strand +

						}//overlap

					}//if there is chromosome in a current chunk

					in >> id;
				}//while
				in.close(); in.clear();

			}//for y all files with results
cout << "Finished calculating ACGT-count from paired-end reads." << endl;
		}//if pair end
	

		if(is_singles == true){
cout << "Processing Single-End results" << endl;
			for( y = 0; y < names_singles2.size(); y++){


				in.open(names_singles2[y].c_str(), ios::in);
				if(!in){
					cerr << "ERROR: can't open " << names_singles2[y] << endl;
					exit(0);
				}
				else
					cout << "opening " << names_singles2[y] << endl;
				long int count_lines = 0;
				in >> id;
				while(!in.eof()){
					in >> read1 ;
					in >> chrom_name; //starts with 1
					in >> strand;
					in >> pos1;
					in >> mism;

					map<string, Point>::iterator afind = ref_seq[t].find(chrom_name);

					if(afind != end){

						chrom_id = (afind->second).x ;
						asize = (afind->second).y;

						if(chrom_id < 0 || chrom_id >= chrom_num)
						{
						cout << "chrom_id is out of boundary at line " << count_lines << endl;
						exit(0);

						}			
						long int len_mer = read1.length();
						j = 0;

						long int forw_pos = pos1;
						long int rev_pos = pos1;
						string forw_read = read1;
						string rev_read = read1;

				
						if(strand == "+"){

							for(i = forw_pos; i < forw_pos + len_mer; i++){

								if(i >= 0 && i < asize){

									if(forw_read[j] == 'A' || forw_read[j] == 'a')
										genome[chrom_id][i].a++;
									else if(forw_read[j] == 'C' || forw_read[j] == 'c')
										genome[chrom_id][i].c++;
									else if(forw_read[j] == 'G' || forw_read[j] == 'g')
										genome[chrom_id][i].g++;
									else if(forw_read[j] == 'T' || forw_read[j] == 't')
										genome[chrom_id][i].t++;
								}//

								j++;
							}//for i
						}else{

							//process reverse read

							//NOTE: we count acgt as they map forward strand,
							//so if we need to use this count for reverse strand, 
							//we have to switch counts of g and c AND counts of t and a
							j=len_mer -1;
							for(i = rev_pos; i < rev_pos + len_mer; i++){
								if(i >= 0 && i < asize){

									if(rev_read[j] == 'A' || rev_read[j] == 'a')
										rev_genome[chrom_id][i].t++;
									else if(rev_read[j] == 'C' || rev_read[j] == 'c')
										rev_genome[chrom_id][i].g++;
									else if(rev_read[j] == 'G' || rev_read[j] == 'g')
										rev_genome[chrom_id][i].c++;
									else if(rev_read[j] == 'T' || rev_read[j] == 't')
										rev_genome[chrom_id][i].a++;
								}//
								j--;
							}//for i
						}//else reverse

					}//if there is chrom in this chunk

					in >> id;
				}//while
				in.close(); in.clear();

			}//for y all files with results
cout << "Finished calculating ACGT-count from single reads." << endl;
		}//if singles
		ofstream out;

		if(is_pe == true || is_singles == true){

			start = ref_seq[t].begin();

			for(; start != end; start++){
				string output =  pref + "_forw_" + start->first + ".txt";
				long int ind = (start->second).x;
				long int asize_of_chrom = (start->second).y;

				out.open(output.c_str(), ios::out);
cout << "Outputting " << output << endl;				
				for(j = 0; j < asize_of_chrom; j++){
					out << genome[ind][j].a << "\t" << genome[ind][j].c << "\t" << genome[ind][j].g << "\t" << genome[ind][j].t << endl;
				}//for j
			out.close(); out.clear();
			}//for i

			start = ref_seq[t].begin();

			for(; start != end; start++){
				string output =  pref + "_rev_" + start->first + ".txt";
				long int ind = (start->second).x;
				long int asize_of_chrom = (start->second).y;

				out.open(output.c_str(), ios::out);
				
				for(j = 0; j < asize_of_chrom; j++){
					out << rev_genome[ind][j].a << "\t" << rev_genome[ind][j].c << "\t" << rev_genome[ind][j].g << "\t" << rev_genome[ind][j].t << endl;
				}//for j
			out.close(); out.clear();
			}//for i
		}//if any input



	}//for ref_seq size

	return 0;

}
